import Fau_tools.torch_tools
from .Fau import *

# __all__ = ["calc_time"]


# import Fau_tools.torch_tools   is for calling function in [Fau_tools.torch_tools.xxx()].
# from .Fau import *   is for calling function in Fau model in [Fau_tools.xxx()]
